;; Pinaret - Static website generator
;; Copyright (C) 2016  Josep Portella Florit <hello@josep-portella.com>
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

#lang racket

(require "provide.rkt" "runtime.rkt")

(struct position/value (position value) #:transparent)
(provide position/value position/value-position position/value-value
         position/value?)
(struct application (procedure arguments) #:transparent)
(provide application application-procedure application-arguments
         application?)

(define/provide escape-char (make-parameter #\\))
(define/provide interpolation-char (make-parameter #\$))
(define/provide application-start-char (make-parameter #\{))
(define/provide application-end-char (make-parameter #\}))
(define/provide quote-chars (make-parameter '(#\' #\" #\`)))
(define/provide filter-char (make-parameter #\|))

(define (concatenate-strings val)
  (if (pair? val)
      (let ([x (first val)]
            [val* (concatenate-strings (rest val))])
        (if (and (string? x) (not (null? val*)) (string? (first val*)))
            (cons (string-append x (first val*)) (rest val*))
            (cons x val*)))
      val))

(define (simplify val)
  (cond [(null? val) ""]
        [(null? (rest val)) (first val)]
        [else val]))

(define (cons-output-string val output)
  (let ([str (get-output-string output)])
    (if (equal? "" str) val (cons str val))))

(define/provide (parse-procedure str [path #f] [line 1] [column 0])
  (call-with-input-string
   str
   (lambda (input)
     (define (position*)
       (let-values ([(nline ncolumn _)
                     (port-next-location input)])
         (position path (+ line (sub1 nline))
                   (if (eqv? 1 nline) (+ column ncolumn) ncolumn))))
     (define (maybe-escape c output)
       (write-char (if (eqv? c (escape-char))
                       (let ([d (read-char input)])
                         (if (eof-object? d)
                             (raise-error/params
                              'unfinished-escape-sequence (position*))
                             d))
                       c)
                   output))
     (define (parse-quoting quote-char beginning)
       (let ([output (open-output-string)])
         (let loop ()
           (let ([c (read-char input)])
             (cond [(eof-object? c)
                    (raise-error/params
                     'unfinished-quoting (position*)
                     'beginning beginning)]
                   [(eqv? (application-start-char) c)
                    (cons-output-string
                     (cons (parse-application)
                           (parse-quoting quote-char beginning))
                     output)]
                   [(eqv? (application-end-char) c)
                    (raise-error/params
                     'unexpected-end-of-procedure-application
                     (position*)
                     'unfinished-quoting-beginning beginning)]
                   [(eqv? quote-char c)
                    (cons-output-string '() output)]
                   [else (maybe-escape c output)
                         (loop)])))))
     (define (parse-values* beginning)
       (let ([output (open-output-string)])
         (define (end-of-values filter?)
           (values (let ([val (cons-output-string '() output)])
                     (if (null? val)
                         '()
                         (list val)))
                   filter?))
         (define (continue vals)
           (call-with-values (λ () (parse-values* beginning))
             (lambda (vals* filter?)
               (values (if (null? vals*)
                           (list (cons-output-string vals output))
                           (cons (cons-output-string
                                  (append vals (first vals*)) output)
                                 (rest vals*)))
                       filter?))))
         (let loop ()
           (let ([c (read-char input)])
             (cond [(eof-object? c)
                    (raise-error/params
                     'unfinished-procedure-application (position*)
                     'beginning beginning)]
                   [(char-whitespace? c)
                    (call-with-values (λ () (parse-values beginning))
                      (lambda (vals filter?)
                        (values (cons (cons-output-string '() output) vals)
                                filter?)))]
                   [(eqv? (application-start-char) c)
                    (continue (list  (parse-application)))]
                   [(eqv? (application-end-char) c) (end-of-values #f)]
                   [(eqv? (filter-char) c) (end-of-values #t)]
                   [(member c (quote-chars))
                    (continue (parse-quoting c (position*)))]
                   [else (maybe-escape c output)
                         (loop)])))))
     (define (skip-whitespace)
       (let ([c (peek-char input)])
         (when (and (not (eof-object? c)) (char-whitespace? c))
           (read-char input)
           (skip-whitespace))))
     (define (parse-values beginning)
       (skip-whitespace)
       (let ([pos (position*)])
         (call-with-values (λ () (parse-values* beginning))
           (lambda (vals filter?)
             (values (if (null? vals)
                         '()
                         (cons (position/value pos
                                               (simplify
                                                (concatenate-strings
                                                 (first vals))))
                               (rest vals)))
                     filter?)))))
     (define (parse-application)
       (let loop ([previous #f])
         (let ([pos (position*)])
           (call-with-values (λ () (parse-values pos))
             (lambda (vals filter?)
               (when (null? vals)
                 (raise-error/params
                  'application-without-procedure pos))
               (let ([app (application (first vals)
                                       (if previous
                                           (cons previous (rest vals))
                                           (rest vals)))])
                 (if filter?
                     (loop (position/value pos app))
                     app)))))))
     (define (outside)
       (let ([output (open-output-string)])
         (let loop ()
           (let ([c (read-char input)])
             (define (escape)
               (write-char c output)
               (let count ([n 0])
                 (let ([e (read-char input)])
                   (cond [(eof-object? e)
                          (raise-error/params 'unfinished-escape-sequence
                                              (position*))]
                         [(eqv? (escape-char) e)
                          (count (add1 n))]
                         [else
                          (do ([m (if (eqv? (application-start-char) e)
                                      n
                                      (add1 n))
                                  (sub1 m)])
                              ((zero? m))
                            (write-char (escape-char) output))
                          (write-char e output)
                          (loop)]))))
             (cond [(eof-object? c) (cons-output-string '() output)]
                   [(eqv? (interpolation-char) c)
                    (let ([d (peek-char input)])
                      (cond [(eof-object? d)
                             (write-char c output)
                             (cons-output-string '() output)]
                            [(eqv? (application-start-char) d)
                             (read-char input)
                             (cons-output-string
                              (cons (parse-application) (outside))
                              output)]
                            [(eqv? (escape-char) d)
                             (read-char input)
                             (escape)]
                            [else (write-char c output)
                                  (loop)]))]
                   [else (write-char c output)
                         (loop)])))))
     (port-count-lines! input)
     (simplify (outside)))))

(define (procedure-symbol-thunk-maker x)
  (if (string? x)
      (let ([sym (string->symbol x)])
        (λ (apply-procedure) (λ () sym)))
      (let ([thunk-maker (object->thunk-maker x)])
        (lambda (apply-procedure)
          (λ () (string->symbol ((thunk-maker apply-procedure))))))))

(define (argument->thunk-maker pos/val)
  (lambda (apply-procedure)
    (procedure/properties
     ((object->thunk-maker (position/value-value pos/val))
      apply-procedure)
     (position/value-position pos/val))))

(define (application->thunk-maker app)
  (let* ([pos/val (application-procedure app)]
         [pos (position/value-position pos/val)]
         [procedure-maker (procedure-symbol-thunk-maker
                           (position/value-value pos/val))]
         [argument-makers (map argument->thunk-maker
                               (application-arguments app))])
    (lambda (apply-procedure)
      (let ([procedure (procedure-maker apply-procedure)]
            [arguments (map (lambda (argument-maker)
                              (argument-maker apply-procedure))
                            argument-makers)])
        (procedure/properties
         (lambda ()
           (parameterize ([current-position pos])
             (apply-procedure (procedure) arguments)))
         pos)))))

(define (object->thunk-maker x)
  (cond [(string? x) (λ (apply-procedure) (λ () x))]
        [(pair? x)
         (let ([thunk-makers (map object->thunk-maker x)])
           (lambda (apply-procedure)
             (let ([thunks (map (lambda (thunk-maker)
                                  (thunk-maker apply-procedure))
                                thunk-makers)])
               (lambda ()
                 (call-with-output-string
                  (lambda (port)
                    (for ([thunk thunks])
                      (display (thunk) port))))))))]
        [(application? x) (application->thunk-maker x)]
        [else (raise-error/params 'wrong-type x)]))

(define/provide (string->thunk-maker str [path #f] [line 1] [column 0])
  (object->thunk-maker (parse-procedure str path line column)))

(define/provide (thunk-maker->thunk thunk-maker)
  (thunk-maker (lambda (sym arguments)
                 (apply call-procedure* (procedure-ref/raise sym)
                        arguments))))
