;; Pinaret - Static website generator
;; Copyright (C) 2016  Josep Portella Florit <hello@josep-portella.com>
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

#lang racket

(require "provide.rkt")

(struct stack (procedure position previous) #:transparent)
(struct procedure-properties (position name doc) #:transparent)

(struct error/params exn:fail (params args) #:transparent)
(provide error/params-params error/params-args error/params?)

(define/provide current-stack (make-parameter '()))
(define/provide current-recursion-limit (make-parameter 1024))
(define/provide current-procedures (make-parameter '()))
(define/provide current-position (make-parameter #f))

(define procedure-properties-hash (make-weak-hasheq))

(define-for-syntax position list)
(define/provide (position path line column)
  (list path line column))

(define/provide position? (list/c any/c any/c any/c))

(define/provide (position->string pos)
  (apply format "~a:~a:~a" (map (λ (x) (if x x "?")) pos)))

(define-for-syntax (position-syntax x)
  (datum->syntax x `',(position (syntax-source x) (syntax-line x)
                                (syntax-column x))))

(define/provide (procedure/properties procedure pos [name #f] [doc ""])
  (hash-set! procedure-properties-hash procedure
             (procedure-properties pos name doc))
  procedure)

(define (procedure-property accessor)
  (lambda (procedure)
    (cond [(hash-ref procedure-properties-hash procedure #f)
           => accessor]
          [else #f])))

(define/provide procedure-position
  (procedure-property procedure-properties-position))
(define/provide procedure-name
  (procedure-property procedure-properties-name))
(define/provide procedure-doc
  (procedure-property procedure-properties-doc))

(define-syntax/provide define-procedure-group
  (syntax-rules ()
    [(_ sym modifier)
     (begin (define sym (hasheq))
            (define-syntax (modifier x)
              (define (helper key doc value)
                #`(set! sym (hash-set sym '#,key
                                      (procedure/properties
                                       #,value #,(position-syntax x)
                                       '#,key #,doc))))
              (syntax-case x ()
                [(_ key value)
                 (helper #'key #'"" #'value)]
                [(_ key doc value)
                 (helper #'key #'doc #'value)])))]))

(define-syntax/provide define-procedure-group/provide
  (syntax-rules ()
    [(_ sym modifier)
     (begin (define-procedure-group sym modifier)
            (provide sym))]))

(define-syntax/provide use-procedure-group
  (syntax-rules ()
    [(_ sym) (current-procedures (cons sym (current-procedures)))]
    [(_ module sym) (begin (require (only-in module sym))
                           (use-procedure-group sym))]))

(define/provide (raise-error/params . args)
  (raise (error/params "" (current-continuation-marks)
                       (current-parameterization) args)))

(define/provide (error/params-raiser . args)
  (lambda (e)
    (raise (if (error/params? e)
               (error/params (exn-message e) (exn-continuation-marks e)
                              (error/params-params e)
                              (append args (error/params-args e)))
               (apply raise-error/params (append args (list e)))))))

(define/provide (print-error/params e)
  (define (p . args) (display (apply format args) (current-error-port)))
  (define (details args [spaces "  "])
    (cond [(exn? (first args))
           (p "~a~a~%" spaces (string-replace (exn-message (first args))
                                              "\n"
                                              (string-append "\n" spaces)))
           (unless (null? (rest args))
             (details (rest args) spaces))]
          [else (p "~a~a" spaces (first args))
                (let ([args* (rest args)])
                  (cond [(null? args*) (p "~%")]
                        [else (p " ~a~%" (let ([x (first args*)])
                                           (if (position? x)
                                               (position->string x)
                                               x)))
                              (cond [(null? (rest args*))]
                                    [else (details (rest args*)
                                                   "    ")])]))]))
  (define (loop s)
    (unless (null? s)
      (loop (stack-previous s))
      (details (list 'procedure (procedure-name (stack-procedure s))
                     'called-from (stack-position s)
                     'defined-in (procedure-position (stack-procedure s))))))
  (call-with-parameterization
   (error/params-params e)
   (lambda ()
     (p "~%")
     (unless (null? (current-stack))
       (p "BACKTRACE:~%")
       (loop (current-stack)))
     (p "ERROR:~%")
     (details (error/params-args e))
     (p "~%"))))

(define-syntax/provide (with-current-position x)
  (syntax-case x ()
    [(_ exp)
     #`(parameterize ([current-position #,(position-syntax x)])
         exp)]))

(define/provide (procedure-ref/forgive sym)
  (let loop ([hashes (current-procedures)])
    (cond [(null? hashes) #f]
          [(hash-ref (first hashes) sym #f)]
          [else (loop (rest hashes))])))

(define/provide (procedure-ref/raise sym)
  (cond [(procedure-ref/forgive sym)]
        [else (raise-error/params 'unknown-procedure sym
                                   'referenced-in (current-position))]))

(define-syntax/provide procedure-ref
  (syntax-rules ()
    [(_ sym) (with-current-position (procedure-ref/raise sym))]))

(define-syntax/provide procedure-ref/call
  (syntax-rules ()
    [(_ sym args ...)
     (call-procedure (procedure-ref/raise sym) args ...)]))

(define-syntax/provide procedure-ref/apply
  (syntax-rules ()
    [(_ sym args ...)
     (apply-procedure (procedure-ref/raise sym) args ...)]))

(define-syntax/provide call-procedure
  (syntax-rules ()
    [(_ proc args ...)
     (with-current-position (call-procedure* proc args ...))]))

(define-syntax/provide apply-procedure
  (syntax-rules ()
    [(_ proc args ...)
     (with-current-position (apply call-procedure* proc args ...))]))

(define/provide (call-procedure* proc . args)
  (let ([n (current-recursion-limit)])
    (if (zero? n)
        (raise-error/params 'reached-recursion-limit)
        (parameterize ([current-recursion-limit (sub1 n)]
                       [current-stack (stack proc (current-position)
                                             (current-stack))])
          (if (arity-includes? (procedure-arity proc) (length args))
              (with-handlers ([exn:fail? (error/params-raiser)])
                (apply proc args))
              (raise-error/params 'wrong-number-of-arguments))))))
