;; Pinaret - Static website generator
;; Copyright (C) 2016  Josep Portella Florit <hello@josep-portella.com>
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

#lang racket

(require "provide.rkt" "runtime.rkt" "base.rkt" "dict.rkt" "repl.rkt")

(struct cache ([library #:mutable]
               library-timestamps
               static-timestamps
               generators
               generators-timestamps))

(define/provide current-cache (make-parameter #f))
(define/provide loop? (make-parameter #f))
(define/provide quiet? (make-parameter #f))
(define/provide dummy? (make-parameter #f))
(define/provide repl-on-error? (make-parameter #f))
(define/provide omit-file? (make-parameter (const #f)))
(define/provide ignore-file? (make-parameter (const #f)))
(define/provide dict-suffix (make-parameter ".dict"))
(define/provide library-dir (make-parameter "library"))
(define/provide input-dir (make-parameter "input"))
(define/provide input-key (make-parameter 'input))
(define/provide path-key (make-parameter 'path))
(define/provide output-dir (make-parameter "output"))
(define/provide index-file-key (make-parameter 'index-file))
(define/provide output-key (make-parameter 'output))
(define/provide bytes?-key (make-parameter 'bytes?))
(define/provide mode-key (make-parameter 'mode))
(define current-generator (make-parameter #f))
(define current-write (make-parameter #f))

(define-procedure-group/provide system-procedures define-system-procedure)

(define/provide (make-cache)
  (cache #f (make-hash) (make-hash) (make-hasheq) (make-hash)))

(define/provide (for-each-file proc path #:max-depth [max-depth 0]
                               #:timestamp? [ts? #f])
  (define (helper trail depth)
    (let ([path/trail (if trail (build-path path trail) path)])
      (for ([entry (with-handlers ([exn:fail? (error/params-raiser)])
                     (directory-list path/trail))])
        (let* ([trail/entry (if trail
                                (build-path trail entry)
                                entry)]
               [path/trail/entry (build-path path trail/entry)])
          (cond [(directory-exists? path/trail/entry)
                 (when (or (zero? max-depth)
                           (< depth max-depth))
                   (helper trail/entry (add1 depth)))]
                [(file-exists? path/trail/entry)
                 (if ts?
                     (proc trail/entry
                           (file-or-directory-modify-seconds
                            path/trail/entry #f (const 0)))
                     (proc trail/entry))])))))
  (helper #f 1))

(define/provide (dict-file? path)
  (equal? path (path-replace-suffix path (dict-suffix))))

(define (read-dict-directory directory)
  (let ([hsh (hasheq)])
    (for-each-file (lambda (path)
                     (when (dict-file? path)
                       (set! hsh
                             (read-dict-file (build-path directory path) hsh
                                             #:thunk-maker->procedure
                                             thunk-maker->procedure))))
                   directory)
    hsh))

(define/provide (library-ref)
  (cond [(cache-library (current-cache))]
        [else (let ([hsh (if (directory-exists? (library-dir))
                             (read-dict-directory (library-dir))
                             (hasheq))])
                (set-cache-library! (current-cache) hsh)
                hsh)]))

(define/provide (relative-path path)
  (cond [(absolute-path? path)
         (let ([subs (rest (explode-path path))])
           (if (null? subs) "." (apply build-path subs)))]
        [(equal? "" path) "."]
        [else path]))

(define/provide (build-path* base . subs)
  (apply build-path (if (equal? "" base) "." base)
         (map relative-path subs)))

(define/provide (symbol->dict-path sym)
  (let ([str (symbol->string sym)])
    (if (equal? "" str)
        (dict-suffix)
        (string-append str (dict-suffix)))))

(define/provide (generator-ref sym)
  (cond [(hash-ref (cache-generators (current-cache)) sym #f)]
        [else (let ([path (build-path* (input-dir) (symbol->dict-path sym))])
                (if (file-exists? path)
                    (let ([hsh (read-dict-file path #:thunk-maker->procedure
                                               thunk-maker->procedure)])
                      (hash-set! (cache-generators (current-cache)) sym hsh)
                      hsh)
                    (raise-error/params 'unknown-generator sym)))]))

(define/provide (call-with-generator sym thunk)
  (parameterize ([current-procedures (list* (generator-ref sym)
                                            (library-ref)
                                            (current-procedures))]
                 [current-generator sym])
    (thunk)))

(define/provide (safe-path path)
  (if (and (not (equal? "" path)) (member 'up (explode-path path)))
      (raise-error/params 'directory-traversal path)
      path))

(define/provide (dict-path->symbol path #:remove-suffix? [r-s? #t])
  (let ([subs (filter (λ (sub) (not (eq? sub 'same)))
                       (explode-path (relative-path path)))])
    (if (null? subs)
        '||
        (string->symbol (path->string (let ([path* (apply build-path subs)])
                                        (if r-s?
                                            (path-replace-suffix path* "")
                                            path*)))))))

(define-system-procedure generator
  "{generator} returns the path of the current generator relative to
   the input directory and without the .dict extension."
  (lambda ()
    (cond [(current-generator) => symbol->string]
          [else (raise-error/params 'unknown)])))

(define/provide (safe-input-dir path)
  (let ([path* (build-path* (input-dir) (safe-path path))])
    (if (directory-exists? path*)
        path*
        (raise-error/params 'unknown-input-directory path))))

(define-system-procedure generators
  "{generators K P D} iterates the generators in the path P (which
   must be relative to the input directory) limited to depth D (where 0
   means no limit); for each generator, temporarily defines a procedure
   named K, and calls {yield}.  {K A} returns A in the context of the
   current generator.  {generators K} is equivalent to {generators K ''
   1}.  {generators K P} is equivalent to {generators K P 1}."
  (lambda (key [path-thunk (const "")] [depth (const "1")])
    (let ([path (path-thunk)]
          [yield (current-yield)])
      (for-each-file
       (lambda (entry)
         (when (dict-file? entry)
           (let ([sym (dict-path->symbol (build-path* path entry))])
             (call-extended (λ () (call-yield yield))
                            key (curry call-with-generator sym)))))
       (safe-input-dir path)
       #:max-depth (string->integer (depth))))
    ""))

(define-system-procedure extern
  "{extern A G} returns A in the context of the generator with path G,
   which must be relative to the input directory, and the .dict extension
   must be omitted.  {extern A G P1 V1 ... Pn Vn} is equivalent to
   {extended {extern A G} P1 V1 ... Pn Vn}."
  (lambda (thunk generator-key . keys/values)
    (call-with-generator (dict-path->symbol (safe-path
                                             (generator-key))
                                            #:remove-suffix? #f)
                         (λ () (call-extended* thunk keys/values)))))

(define-system-procedure input-files
  "{input-files K P D} iterates the files in the path P (which must be
   relative to the input directory) limited to depth D (where 0 means no
   limit); for each file, temporarily defines a procedure named K that
   returns the current file name, and calls {yield}.  {input-files K} is
   equivalent to {input-files K '' 1}.  {input-files K P} is equivalent
   to {input-files K P 1}."
  (lambda (key [path-thunk (const "")] [depth (const "1")])
    (let ([path (path-thunk)]
          [yield (current-yield)])
      (for-each-file
       (lambda (entry)
         (call-extended (λ () (call-yield yield))
                        key (λ () (build-path* path entry))))
       (safe-input-dir path)
       #:max-depth (string->integer (depth))))
    ""))

(define/provide (safe-input-file path)
  (let ([path* (build-path* (input-dir) (safe-path path))])
    (if (file-exists? path*)
        path*
        (raise-error/params 'unknown-input-file path))))

(define-system-procedure input-file-size
  "{input-file-size F} returns the size in bytes of the file with path
   F, which must be relative to the input directory."
  (lambda (path-thunk)
    (number->string (file-size (safe-input-file (path-thunk))))))

(define-system-procedure input-file-string
  "{input-file-string F} returns the contents of the file with path
   F (which must be relative to the input directory) decoded using
   UTF-8."
  (λ (path-thunk) (file->string (safe-input-file (path-thunk)))))

(define-system-procedure input-file-bytes
  "{input-file-bytes F} returns the contents of the file with path
   F (which must be relative to the input directory) decoded using
   Latin-1."
  (lambda (path-thunk)
    (bytes->string/latin-1 (file->bytes (safe-input-file (path-thunk))))))

(define-system-procedure input-file-timestamp
  "{input-file-timestamp F} returns the modification timestamp of the
   file with path F (which must be relative to the input directory) in
   the ISO-8601 year-month-day-hour-minute-second-timezone format.  The
   resulting string reflects the time in UTC, but {input-file-timestamp F
   true} reflects the time according to the local time zone."
  (lambda (path-thunk [local? false-string])
    (date->iso8601-string (seconds->date (file-or-directory-modify-seconds
                                          (safe-input-file (path-thunk)))
                                         (string->boolean (local?))))))

(define-system-procedure write
  "{write A} writes A to the file that is currently being generated."
  (lambda (thunk)
    (cond [(current-write) => (λ (write) (write (thunk)))]
          [else (raise-error/params 'not-generating)])
    ""))

(define/provide (update-cache/generate?)
  (let ([generate? #f])
    (define (record-path path ts paths hsh [proc void])
      (set-add! paths path)
      (unless (= ts (hash-ref hsh path 0))
        (proc)
        (hash-set! hsh path ts)
        (set! generate? #t)))
    (define (check-paths paths hsh [proc void])
      (for ([path (hash-keys hsh)])
        (unless (set-member? paths path)
          (proc path)
          (hash-remove! hsh path)
          (set! generate? #t))))
    (let ([dict-paths (mutable-set)])
      (when (directory-exists? (library-dir))
        (for-each-file (lambda (path ts)
                         (when (dict-file? path)
                           (record-path path ts dict-paths
                                        (cache-library-timestamps
                                         (current-cache)))))
                       (library-dir) #:timestamp? #t))
      (check-paths dict-paths (cache-library-timestamps (current-cache)))
      (when generate? (set-cache-library! (current-cache) #f)))
    (let ([static-paths (mutable-set)]
          [dict-paths (mutable-set)])
      (when (directory-exists? (input-dir))
        (for-each-file
         (lambda (path ts)
           (cond [(dict-file? path)
                  (record-path path ts dict-paths
                               (cache-generators-timestamps (current-cache))
                               (lambda ()
                                 (hash-set! (cache-generators
                                             (current-cache))
                                            (dict-path->symbol path) #f)))]
                 [(not ((ignore-file?) path))
                  (record-path path ts static-paths
                               (cache-static-timestamps (current-cache)))]))
         (input-dir) #:timestamp? #t))
      (check-paths dict-paths (cache-generators-timestamps (current-cache))
                   (lambda (path)
                     (hash-remove! (cache-generators (current-cache))
                                   (dict-path->symbol path))))
      (check-paths static-paths (cache-static-timestamps (current-cache))))
    generate?))


(define (displayln/time v)
  (let ((d (seconds->date (current-seconds))))
    (define (z x)
      (if (< x 10)
          (string-append "0" (number->string x))
          x))
    (printf "~a:~a:~a ~a~%" (z (date-hour d)) (z (date-minute d))
            (z (date-second d)) v)))

(define/provide (generate-variant sym)
  (with-handlers ([exn:fail? (error/params-raiser 'generate-variant sym)])
    (let* ([path (simplify-path
                  (cond [(procedure-ref/forgive (path-key))
                         => (λ (proc) (call-procedure proc))]
                        [else (symbol->string sym)])
                  #f)]
           [path (relative-path
                  (safe-path (if (equal? path (path->directory-path path))
                                 (build-path* path (procedure-ref/call
                                                    (index-file-key)))
                                 path)))]
           [path* (build-path* (output-dir) path)])
      (unless (dummy?) (make-directory* (path-only path*)))
      (let ([port (if (dummy?)
                      (open-output-nowhere)
                      (open-output-file path* #:exists 'truncate/replace))])
        (with-handlers ([exn:fail? (error/params-raiser 'file path*)])
          (dynamic-wind
            void
            (lambda ()
              (define write!
                (if (cond [(procedure-ref/forgive (bytes?-key))
                           => (lambda (proc)
                                (string->boolean (call-procedure proc)))]
                          [else #f])
                    (lambda (str)
                      (display (with-handlers ([exn:fail:contract?
                                                (lambda _
                                                  (raise-error/params
                                                   'encoding-error str))])
                                 (string->bytes/latin-1 str))
                               port))
                    (curryr display port)))
              (cond [(procedure-ref/forgive (mode-key))
                     => (lambda (proc)
                          (let ([mode (string->number*
                                       (call-procedure proc))])
                            (unless (dummy?)
                              (file-or-directory-permissions
                               path* (bitwise-and (abs mode) #o7777)))))])
              (write! (parameterize ([current-write write!])
                        (procedure-ref/call (output-key)))))
            (λ () (close-output-port port)))))
      (unless (quiet?)
        (displayln/time (if (eq? sym (string->symbol (path->string path)))
                            sym
                            (format "~a -> ~a" sym path)))))))

(define/provide (generate-variants sym [input-thunk #f])
  (define (yield) (generate-variant sym))
  (call-with-generator
   sym (lambda ()
         (cond [input-thunk (parameterize ([current-yield yield])
                              (input-thunk))]
               [(procedure-ref/forgive (input-key))
                => (lambda (input-proc)
                     (parameterize ([current-yield yield])
                       (call-procedure input-proc)))]
               [else (generate-variant sym)]))))

(define (older? path ts)
  (> ts (file-or-directory-modify-seconds path #f (const 0))))

(define/provide (copy-static-files)
  (when (and (not (dummy?)) (directory-exists? (input-dir)))
    (for-each-file
     (lambda (path ts)
       (let ([orig (build-path* (input-dir) path)])
         (unless (or (dict-file? orig) ((omit-file?) orig))
           (let ([dest (build-path* (output-dir) path)])
             (when (or (older? dest ts)
                       (with-handlers ([exn:fail? (const #t)])
                         (not (= (file-size orig) (file-size dest)))))
               (with-handlers ([exn:fail? (error/params-raiser)])
                 (make-directory* (path-only dest))
                 (copy-file orig dest #t))
               (unless (quiet?) (displayln/time path)))))))
     (input-dir) #:timestamp? #t)))

(define/provide (generate [generator-paths '()] [input-thunk #f])
  (let ([syms (with-handlers ([error/params? (lambda (e)
                                               (print-error/params e)
                                               (exit 1))])
                (map (lambda (path)
                       (dict-path->symbol (safe-path path)
                                          #:remove-suffix? #f))
                     generator-paths))])
    (let loop ([first? #t])
      (with-handlers ([error/params?
                       (lambda (e)
                         (print-error/params e)
                         (call-with-parameterization
                          (error/params-params e)
                          (lambda ()
                            (unless (not (repl-on-error?))
                              (parameterize ([current-stack '()])
                                (repl))
                              (newline)
                              (newline))))
                         (unless (loop?) (exit 1)))])
        (when (or (update-cache/generate?) first?)
          (copy-static-files)
          (if (null? syms)
              (when (directory-exists? (input-dir))
                (for-each-file
                 (lambda (path)
                   (when (dict-file? path)
                     (generate-variants (dict-path->symbol path)
                                        input-thunk)))
                 (input-dir)))
              (for-each (λ (sym) (generate-variants sym input-thunk))
                        syms))))
      (when (loop?)
        (sleep 1)
        (loop #f)))))
